local tonumber = tonumber
local format = format
local tinsert = tinsert
local ipairs = ipairs
local strfind = strfind
local select = select
local strsplit = strsplit
local wipe = wipe

-- Any slot has a list linked to it, this list might contain 1 item, the best for that slot, or all items of that slot, depending if the max self dps item has hit
local RhadaTip_BestHeadLink = {}
local RhadaTip_BestNeckLink = {}
local RhadaTip_BestShoulderLink = {}
local RhadaTip_BestBackLink = {}
local RhadaTip_BestChestLink = {}
local RhadaTip_BestWristLink = {}
local RhadaTip_BestHandsLink = {}
local RhadaTip_BestWaistLink = {}
local RhadaTip_BestLegsLink = {}
local RhadaTip_BestFeetLink = {}
local RhadaTip_BestFinger0Link = {}
local RhadaTip_BestTrinket0Link = {}
local RhadaTip_BestMainHandLink = {}
local RhadaTip_BestSecondaryHandLink = {}
local RhadaTip_BestRangedLink = {}

RhadaTip_gear_slots = {
"HeadSlot",
"NeckSlot",
"ShoulderSlot",
"BackSlot",
"ChestSlot",
"WristSlot",
"HandsSlot",
"WaistSlot",
"LegsSlot",
"FeetSlot",
"MainHandSlot",
"SecondaryHandSlot",
"RangedSlot",
"Finger0Slot",
"Finger1Slot",
"Trinket0Slot",
"Trinket1Slot"
}

RhadaTip_EquipSlotTranslate = {}
RhadaTip_EquipSlotTranslate["INVTYPE_AMMO"] = "AmmoSlot" 
RhadaTip_EquipSlotTranslate["INVTYPE_CLOAK"] = "BackSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_BAG"] = "Bag0Slot"
RhadaTip_EquipSlotTranslate["INVTYPE_CHEST"] = "ChestSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_ROBE"] = "ChestSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_FEET"] = "FeetSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_FINGER"] = "Finger0Slot"
RhadaTip_EquipSlotTranslate["INVTYPE_HAND"] = "HandsSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_HANDS"] = "HandsSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_HEAD"] = "HeadSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_LEGS"] = "LegsSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_WEAPON"] = "MainHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_2HWEAPON"] = "MainHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_WEAPONMAINHAND"] = "MainHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_NECK"] = "NeckSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_RANGED"] = "RangedSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_THROWN"] = "RangedSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_RANGEDRIGHT"] = "RangedSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_RELIC"] = "RangedSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_SHIELD"] = "SecondaryHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_WEAPONOFFHAND"] = "SecondaryHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_HOLDABLE"] = "SecondaryHandSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_BODY"] = "ShirtSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_SHOULDER"] = "ShoulderSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_TABARD"] = "TabardSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_TRINKET"] = "Trinket0Slot"
RhadaTip_EquipSlotTranslate["INVTYPE_WAIST"] = "WaistSlot"
RhadaTip_EquipSlotTranslate["INVTYPE_WRIST"] = "WristSlot"

RhadaTip_Gear_Table = {}

-- items with an item level lower than this will not be considered
local maxItemLevel

RhadaTip_gear_procs = {}
RhadaTip_gear_procs[52572] = {[1] = {["id"] = 52572, ["proc_len"] = 10, ["proc_chan"] = 1, ["proc_ICD"] = 60,["proc_stat"] = "ITEM_MOD_ATTACK_POWER_SHORT",["stat_val"] = 480}}

-- this function fills the mainhand and offhand lists with all weapons, so weapons arent considered like gloves or belts anymore. depending on the equip location, a weapon can be inserted either in the
-- mainhand list, or the offhand list, or in both (this should ensure that if a staff is worth more than a main hand, it wont be taken over the main hand + off hand combo, as both will be tested, since
-- both the staff and the main hand are put into RhadaTip_BestMainHandLink)
function RhadaTip_init_gear_table()
	wipe(RhadaTip_Gear_Table)
	for k,v in ipairs(RhadaTip_gear_slots) do
		RhadaTip_Gear_Table[v] = GetInventoryItemLink("player",GetInventorySlotInfo(v))
	end
end

local function RhadaTip_FillWeaponSlotList()
	for bag = 0,4 do
    		for slot = 1,GetContainerNumSlots(bag) do
      			local item = GetContainerItemLink(bag,slot)
			if item then
	      			local _, _, _, itemLevel, _, _, _, _, itemEquipLoc = GetItemInfo(item)
				if tonumber(itemLevel) >= maxItemLevel then
					-- if the item location is either weaponmainhand or 2hweapon, put it into RhadaTip_BestMainHandLink. If it is a One-Hand, also put it into the MainHand list
					if (itemEquipLoc == "INVTYPE_WEAPONMAINHAND" or 
						itemEquipLoc == "INVTYPE_2HWEAPON" or 
						itemEquipLoc == "INVTYPE_WEAPON") then
						tinsert(RhadaTip_BestMainHandLink, item)
					end
					-- if the item location is either weaponoffhand or holdable or a shield, put it into RhadaTip_BestMainHandLink. If it is a One-Hand, also put it into the MainHand list
					if ((itemEquipLoc == "INVTYPE_WEAPONOFFHAND" and RhadaTip_class.is_dual_wield()) or
						itemEquipLoc == "INVTYPE_HOLDABLE" or 
						itemEquipLoc == "INVTYPE_SHIELD" or 
						(itemEquipLoc == "INVTYPE_WEAPON" and RhadaTip_class.is_dual_wield())) then
						tinsert(RhadaTip_BestSecondaryHandLink, item)
					end
				end
			end
    		end
  	end
	local headLink = GetInventoryItemLink("player",GetInventorySlotInfo("MainHandSlot"))
	local headLink2 = GetInventoryItemLink("player",GetInventorySlotInfo("SecondaryHandSlot"))
	-- if the player has something equipped in the main hand slot
	if headLink ~= nil then
		-- get the item's id and equip location to put it into the appropriate list
		local _, _, _, itemLevel, _, _, _, _, itemEquipLoc = GetItemInfo(headLink)
		if tonumber(itemLevel) >= maxItemLevel then
			-- if the item in the main hand slot is a main hand or a 2h, put it into RhadaTip_BestMainHandLink
			if (itemEquipLoc == "INVTYPE_WEAPONMAINHAND" or itemEquipLoc == "INVTYPE_2HWEAPON") then
				tinsert(RhadaTip_BestMainHandLink, headLink)
			end
			-- if the item in the main hand slot is a one hand, put it into RhadaTip_BestMainHandLink and RhadaTip_BestSecondaryHandLink
			if (itemEquipLoc == "INVTYPE_WEAPON") then
				tinsert(RhadaTip_BestMainHandLink, headLink)
				tinsert(RhadaTip_BestSecondaryHandLink, headLink)
			end
		end
	end
	-- if the player has something equipped in the secondary hand slot
	if headLink2 ~= nil then
		local _, _, _, itemLevel, _, _, _, _, itemEquipLoc = GetItemInfo(headLink2)
		if tonumber(itemLevel) >= maxItemLevel then
			if (itemEquipLoc == "INVTYPE_WEAPONOFFHAND" or itemEquipLoc == "INVTYPE_HOLDABLE" or itemEquipLoc == "INVTYPE_SHIELD") then
				tinsert(RhadaTip_BestSecondaryHandLink, headLink2)
			end
			if (itemEquipLoc == "INVTYPE_WEAPON") then
				tinsert(RhadaTip_BestMainHandLink, headLink2)
				tinsert(RhadaTip_BestSecondaryHandLink, headLink2)
			end
		end
	end
end

-- this function takes a slot as its input, and puts every single item of that kind (equipped and from bags) into the list linked to the item slot
local function RhadaTip_FillSlotList(slotToCompute)
	local tableSlot, currentSlot
	-- find out which itemList is linked to the slot of interest
	if slotToCompute == "INVTYPE_HEAD" then
		tableSlot=RhadaTip_BestHeadLink
		currentSlot = "HeadSlot"
	elseif slotToCompute == "INVTYPE_NECK" then
		tableSlot=RhadaTip_BestNeckLink
		currentSlot = "NeckSlot"
	elseif slotToCompute == "INVTYPE_SHOULDER" then
		tableSlot=RhadaTip_BestShoulderLink
		currentSlot = "ShoulderSlot"
	elseif slotToCompute == "INVTYPE_CLOAK" then
		tableSlot=RhadaTip_BestBackLink
		currentSlot = "BackSlot"
	elseif (slotToCompute == "INVTYPE_ROBE" or slotToCompute == "INVTYPE_CHEST") then
		tableSlot=RhadaTip_BestChestLink
		currentSlot = "ChestSlot"
	elseif slotToCompute == "INVTYPE_WRIST" then
		tableSlot=RhadaTip_BestWristLink
		currentSlot = "WristSlot"
	elseif slotToCompute == "INVTYPE_HAND" then
		tableSlot=RhadaTip_BestHandsLink
		currentSlot = "HandsSlot"
	elseif slotToCompute == "INVTYPE_WAIST" then
		tableSlot=RhadaTip_BestWaistLink
		currentSlot = "WaistSlot"
	elseif slotToCompute == "INVTYPE_LEGS" then
		tableSlot=RhadaTip_BestLegsLink
		currentSlot = "LegsSlot"
	elseif slotToCompute == "INVTYPE_FEET" then
		tableSlot=RhadaTip_BestFeetLink
		currentSlot = "FeetSlot"
	elseif slotToCompute == "INVTYPE_FINGER" then
		tableSlot=RhadaTip_BestFinger0Link
	elseif slotToCompute == "INVTYPE_TRINKET" then
		tableSlot=RhadaTip_BestTrinket0Link
	elseif (slotToCompute == "INVTYPE_WEAPONMAINHAND" or slotToCompute == "INVTYPE_2HWEAPON")then
		tableSlot=RhadaTip_BestMainHandLink
		currentSlot = "MainHandSlot"
	elseif (slotToCompute == "INVTYPE_WEAPONOFFHAND" or slotToCompute == "INVTYPE_HOLDABLE") then
		tableSlot=RhadaTip_BestSecondaryHandLink
		currentSlot = "SecondaryHandSlot"
	elseif (slotToCompute == "INVTYPE_RANGEDRIGHT" or slotToCompute == "INVTYPE_RANGED" or slotToCompute == "INVTYPE_THROWN") then
		tableSlot=RhadaTip_BestRangedLink
		currentSlot = "RangedSlot"
	end
	-- scan all the bags
	for bag = 0,4 do
    		for slot = 1,GetContainerNumSlots(bag) do
      			local item = GetContainerItemLink(bag,slot)
			if item then
	      			local _, _, _, itemLevel, _, _, _, _, itemEquipLoc = GetItemInfo(item)
				if tonumber(itemLevel) >= maxItemLevel then
					-- if the n-th item is of a slot of interest...
					if (itemEquipLoc == slotToCompute or (slotToCompute == "INVTYPE_ROBE" and itemEquipLoc == "INVTYPE_CHEST") or (slotToCompute == "INVTYPE_WEAPONMAINHAND" and itemEquipLoc == "INVTYPE_2HWEAPON")  or (slotToCompute == "INVTYPE_WEAPONOFFHAND" and itemEquipLoc == "INVTYPE_HOLDABLE") or (slotToCompute == "INVTYPE_RANGEDRIGHT" and (itemEquipLoc == "INVTYPE_RANGED" or itemEquipLoc == "INVTYPE_THROWN"))) then
						-- put it in its corresponding list
						tinsert(tableSlot, item)
					end
				end
			end
    		end
  	end
	-- finally insert the currently equipped one
	if (slotToCompute ~= "INVTYPE_FINGER" and slotToCompute ~= "INVTYPE_TRINKET") then
		-- if the slot of interest isnt a ring or trinket, simply get what the user is wearing into the currentSlot
		local headLink = GetInventoryItemLink("player",GetInventorySlotInfo(currentSlot))
		if headLink ~= nil then
			local itemLevel = select(4, GetItemInfo(headLink))
			if tonumber(itemLevel) >= maxItemLevel then
				tinsert(tableSlot, headLink)
			end
		end
	else
		-- otherwise, if currentSlot is a ring, get both the player rings
		if (slotToCompute == "INVTYPE_FINGER") then
			local headLink = GetInventoryItemLink("player",GetInventorySlotInfo("Finger0Slot"))
			local headLink2 = GetInventoryItemLink("player",GetInventorySlotInfo("Finger1Slot"))
			if headLink ~= nil then
				local itemLevel = select(4, GetItemInfo(headLink))
				if tonumber(itemLevel) >= maxItemLevel then
					tinsert(tableSlot, headLink)
				end
			end
			if headLink2 ~= nil then
				local itemLevel = select(4, GetItemInfo(headLink2))
				if tonumber(itemLevel) >= maxItemLevel then
					tinsert(tableSlot, headLink2)
				end
			end
		elseif (slotToCompute == "INVTYPE_TRINKET") then
			local headLink = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket0Slot"))
			local headLink2 = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket1Slot"))
			if headLink ~= nil then
				local itemLevel = select(4, GetItemInfo(headLink))
				if tonumber(itemLevel) >= maxItemLevel then
					tinsert(tableSlot, headLink)
				end
			end
			if headLink2 ~= nil then
				local itemLevel = select(4, GetItemInfo(headLink2))
				if tonumber(itemLevel) >= maxItemLevel then
					tinsert(tableSlot, headLink2)
				end
			end
		end
	end	
end

-- this function clears all the ITEM_SLOT lists
local function RhadaTip_ClearGearLists()
	wipe(RhadaTip_BestHeadLink)
	wipe(RhadaTip_BestNeckLink)
	wipe(RhadaTip_BestShoulderLink)
	wipe(RhadaTip_BestBackLink)
	wipe(RhadaTip_BestChestLink)
	wipe(RhadaTip_BestWristLink)
	wipe(RhadaTip_BestHandsLink)
	wipe(RhadaTip_BestWaistLink)
	wipe(RhadaTip_BestLegsLink)
	wipe(RhadaTip_BestFeetLink)
	wipe(RhadaTip_BestFinger0Link)
	wipe(RhadaTip_BestTrinket0Link)
	wipe(RhadaTip_BestMainHandLink)
	wipe(RhadaTip_BestSecondaryHandLink)
	wipe(RhadaTip_BestRangedLink)
	--RhadaTip_BestHeadLink = {}
	--RhadaTip_BestNeckLink = {}
	--RhadaTip_BestShoulderLink = {}
	--RhadaTip_BestBackLink = {}
	--RhadaTip_BestChestLink = {}
	--RhadaTip_BestWristLink = {}
	--RhadaTip_BestHandsLink = {}
	--RhadaTip_BestWaistLink = {}
	--RhadaTip_BestLegsLink = {}
	--RhadaTip_BestFeetLink = {}
	--RhadaTip_BestFinger0Link = {}
	--RhadaTip_BestTrinket0Link = {}
	--RhadaTip_BestMainHandLink = {}
	--RhadaTip_BestSecondaryHandLink = {}
	--RhadaTip_BestRangedLink = {}
end

-- this function returns the number of possible gear setups obtainable with the current gear plus the items in the bags
local function RhadaTip_GetGearIterationNumber()
	local nthHead, nthNeck, nthShoulder, nthBack, nthChest, nthWrist, nthHands, nthWaist, nthLegs, nthFeet, nthMainHand, nthSecondaryHand, nthRanged, nthFinger0, nthFinger1, nthTrinket0, nthTrinket1;
	local loopNesimo = 0;
	for k,v in ipairs(RhadaTip_BestHeadLink) do
		nthHead = v
		for k,v in ipairs(RhadaTip_BestNeckLink) do
			nthNeck = v
			for k,v in ipairs(RhadaTip_BestShoulderLink) do
				nthShoulder = v
				for k,v in ipairs(RhadaTip_BestBackLink) do
					nthBack = v
					for k,v in ipairs(RhadaTip_BestChestLink) do
						nthChest = v
						for k,v in ipairs(RhadaTip_BestWristLink) do
							nthWrist = v
							for k,v in ipairs(RhadaTip_BestHandsLink) do
								nthHands = v
								for k,v in ipairs(RhadaTip_BestWaistLink) do
									nthWaist = v
									for k,v in ipairs(RhadaTip_BestLegsLink) do
										nthLegs = v
										for k,v in ipairs(RhadaTip_BestFeetLink) do
											nthFeet = v
											for k,v in ipairs(RhadaTip_BestMainHandLink) do
												nthMainHand = v
												local ItemEquipLoc
												if nthMainHand ~= "MainHandSlot=nil" then
													local ItemID = tonumber(select(3, strfind(nthMainHand, "Hitem:(%d+)")))
													ItemEquipLoc = select(9, GetItemInfo(ItemID))
												end
												for k,v in ipairs(RhadaTip_BestSecondaryHandLink) do
													if ((v == "SecondaryHandSlot=nil") or (ItemEquipLoc == nil) or (ItemEquipLoc ~= "INVTYPE_2HWEAPON" and ItemEquipLoc ~= "INVTYPE_WEAPON")) then
														nthSecondaryHand = v
													else
														if (ItemEquipLoc == "INVTYPE_2HWEAPON") then
															nthSecondaryHand = "SecondaryHandSlot=nil"
														else 
															if (ItemEquipLoc == "INVTYPE_WEAPON") then
																local ItemIdOh = tonumber(select(3, strfind(v, "Hitem:(%d+)")))
																if ItemID == ItemIdOh then
																	local itemCount = GetItemCount(ItemID) 
																	if itemCount > 1 then
																		-- the user actually has more than 1 of this item, so count both
																		nthSecondaryHand = v
																	else
																		nthSecondaryHand = "SecondaryHandSlot=nil"
																	end
																else
																	nthSecondaryHand = v
																end
															end
														end
													end
													for k,v in ipairs(RhadaTip_BestRangedLink) do
														nthRanged = v
														for i=1,(table.maxn(RhadaTip_BestFinger0Link)-1) do
															local counter = i
															nthFinger0 = RhadaTip_BestFinger0Link[i]
															for i=(1+counter),table.maxn(RhadaTip_BestFinger0Link) do
																nthFinger1 = RhadaTip_BestFinger0Link[i]
																for i=1,(table.maxn(RhadaTip_BestTrinket0Link)-1) do
																	local counterTrinket = i
																	nthTrinket0 = RhadaTip_BestTrinket0Link[i]
																	for i=(1+counterTrinket),table.maxn(RhadaTip_BestTrinket0Link) do
																		nthTrinket1 = RhadaTip_BestTrinket0Link[i]
																		loopNesimo = loopNesimo + 1;
																	end
																end
															end
														end
													end
												end
											end
										end
									end
								end
							end
						end
					end
				end
			end
		end
	end
	-- once the exact number of gear combination is known, print a message to let the user know that the computation has been fired
	Rhada_config_panel_gear_opt_ComputationTime:SetText("RhadaTip's started to evaluate all the " .. loopNesimo .. " possible gear combinations, please wait ...")
end

-- if one of the item lists is empty, insert a fictional element in it
local function RhadaTip_FillAllGearSetsLists()
	if (RhadaTip_BestHeadLink[1] == nil) then
		tinsert(RhadaTip_BestHeadLink, "HeadSlot=nil")
	end
	if (RhadaTip_BestNeckLink[1] == nil) then
		tinsert(RhadaTip_BestNeckLink, "NeckSlot=nil")
	end
	if (RhadaTip_BestShoulderLink[1] == nil) then
		tinsert(RhadaTip_BestShoulderLink, "ShoulderSlot=nil")
	end
	if (RhadaTip_BestBackLink[1] == nil) then
		tinsert(RhadaTip_BestBackLink, "BackSlot=nil")
	end
	if (RhadaTip_BestChestLink[1] == nil) then
		tinsert(RhadaTip_BestChestLink, "ChestSlot=nil")
	end
	if (RhadaTip_BestWristLink[1] == nil) then
		tinsert(RhadaTip_BestWristLink, "WristSlot=nil")
	end
	if (RhadaTip_BestHandsLink[1] == nil) then
		tinsert(RhadaTip_BestHandsLink, "HandsSlot=nil")
	end
	if (RhadaTip_BestWaistLink[1] == nil) then
		tinsert(RhadaTip_BestWaistLink, "WaistSlot=nil")
	end
	if (RhadaTip_BestLegsLink[1] == nil) then
		tinsert(RhadaTip_BestLegsLink, "LegsSlot=nil")
	end
	if (RhadaTip_BestFeetLink[1] == nil) then
		tinsert(RhadaTip_BestFeetLink, "FeetSlot=nil")
	end
	-- the finger and trinket lists must have at least 2 elements in them, check both [1] and [2]
	if (RhadaTip_BestFinger0Link[1] == nil) then
		tinsert(RhadaTip_BestFinger0Link, "Finger0Slot=nil")
	end
	if (RhadaTip_BestFinger0Link[2] == nil) then
		tinsert(RhadaTip_BestFinger0Link, "Finger1Slot=nil")
	end
	if (RhadaTip_BestTrinket0Link[1] == nil) then
		tinsert(RhadaTip_BestTrinket0Link, "Trinket0Slot=nil")
	end
	if (RhadaTip_BestTrinket0Link[2] == nil) then
		tinsert(RhadaTip_BestTrinket0Link, "Trinket1Slot=nil")
	end
	if (RhadaTip_BestMainHandLink[1] == nil) then
		tinsert(RhadaTip_BestMainHandLink, "MainHandSlot=nil")
	end
	if (RhadaTip_BestSecondaryHandLink[1] == nil) then
		tinsert(RhadaTip_BestSecondaryHandLink, "SecondaryHandSlot=nil")
	end
	if (RhadaTip_BestRangedLink[1] == nil) then
		tinsert(RhadaTip_BestRangedLink, "RangedSlot=nil")
	end	
end

-- this function uses the lists linked to each slot to build all the possible gear combinations that will later be tested to see which one delivers the highest dps
local function RhadaTip_PrintAllGearSetsLists(slideValue)
	local dps, topDps = 0, 0
	-- string containing all the itemLinks that deliver the highest dps
	local topDpsSetup = ""
	-- if one of the item lists is empty, the loop below will break, insert a fictional element into the ones whom list[1] is nil
	RhadaTip_FillAllGearSetsLists();
	-- i need variables to carry the nth item to the underlying loops
	local nthHead, nthNeck, nthShoulder, nthBack, nthChest, nthWrist, nthHands, nthWaist, nthLegs, nthFeet, nthMainHand, nthSecondaryHand, nthRanged, nthFinger0, nthFinger1, nthTrinket0, nthTrinket1;
	RhadaTip_GetGearIterationNumber();
	-- used in order to keep track of the n-th loop (to manage the coroutine)
	local loopNesimo = 0;
	for k,v in ipairs(RhadaTip_BestHeadLink) do
		nthHead = v
		for k,v in ipairs(RhadaTip_BestNeckLink) do
			nthNeck = v
			for k,v in ipairs(RhadaTip_BestShoulderLink) do
				nthShoulder = v
				for k,v in ipairs(RhadaTip_BestBackLink) do
					nthBack = v
					for k,v in ipairs(RhadaTip_BestChestLink) do
						nthChest = v
						for k,v in ipairs(RhadaTip_BestWristLink) do
							nthWrist = v
							for k,v in ipairs(RhadaTip_BestHandsLink) do
								nthHands = v
								for k,v in ipairs(RhadaTip_BestWaistLink) do
									nthWaist = v
									for k,v in ipairs(RhadaTip_BestLegsLink) do
										nthLegs = v
										for k,v in ipairs(RhadaTip_BestFeetLink) do
											nthFeet = v
											for k,v in ipairs(RhadaTip_BestMainHandLink) do
												nthMainHand = v
												local ItemEquipLoc
												if nthMainHand ~= "MainHandSlot=nil" then
													local ItemID = tonumber(select(3, strfind(nthMainHand, "Hitem:(%d+)")))
													ItemEquipLoc = select(9, GetItemInfo(ItemID))
												end
												-- the off hand loop is special. if the current main hand is a 2h weapon, we will set the off hand weapon to the fictional value in the next loop as 
												-- it must not be considered
												for k,v in ipairs(RhadaTip_BestSecondaryHandLink) do
													-- if the current main hand is not a 2h or a one hand, it isnt nil nor is the off hand nil, set nthSecondaryHand normally to v
													if ((v == "SecondaryHandSlot=nil") or (ItemEquipLoc == nil) or (ItemEquipLoc ~= "INVTYPE_2HWEAPON" and ItemEquipLoc ~= "INVTYPE_WEAPON")) then
														nthSecondaryHand = v
													else
														-- otherwise, the current main hand is either a 2h or a one-hand, check which case it is
														if (ItemEquipLoc == "INVTYPE_2HWEAPON") then
															-- if the current main hand is a 2h, set the secondary hand to the fictional value
															nthSecondaryHand = "SecondaryHandSlot=nil"
														else 
															if (ItemEquipLoc == "INVTYPE_WEAPON") then
																--local ItemIdOh = tonumber(v:match("Hitem:(%d+)"))
																local ItemIdOh = tonumber(select(3, strfind(v, "Hitem:(%d+)")))
																if ItemID == ItemIdOh then
																	local itemCount = GetItemCount(ItemID) 
																	if itemCount > 1 then
																		-- the user actually has more than 1 of this item, so count both
																		nthSecondaryHand = v
																	else
																		nthSecondaryHand = "SecondaryHandSlot=nil"
																	end
																else
																	nthSecondaryHand = v
																end
															end
														end
													end
													for k,v in ipairs(RhadaTip_BestRangedLink) do
														nthRanged = v
														-- like this i have Permutations with repetition, 32, i want Combinations without repetition though
														for i=1,(table.maxn(RhadaTip_BestFinger0Link)-1) do
															local counter = i
															nthFinger0 = RhadaTip_BestFinger0Link[i]
															for i=(1+counter),table.maxn(RhadaTip_BestFinger0Link) do
																nthFinger1 = RhadaTip_BestFinger0Link[i]
																for i=1,(table.maxn(RhadaTip_BestTrinket0Link)-1) do
																	local counterTrinket = i
																	nthTrinket0 = RhadaTip_BestTrinket0Link[i]
																	for i=(1+counterTrinket),table.maxn(RhadaTip_BestTrinket0Link) do
																		nthTrinket1 = RhadaTip_BestTrinket0Link[i]
																		--RhadaTip:debug_out(nthHead .. "+" .. nthNeck .. "+" .. nthShoulder .. "+" .. nthBack .. "+" .. nthChest .. "+" .. nthWrist .. "+" .. nthHands .. "+" .. nthWaist .. "+" .. nthLegs .. "+" .. nthFeet .. "+" .. nthMainHand .. "+" .. nthSecondaryHand .. "+" .. nthRanged .. "+" .. nthFinger0 .. "+" .. nthFinger1 .. "+" .. nthTrinket0 .. "+" .. nthTrinket1)
																		-- once i have all the current items for each slot, fire a method to compute the gear value	
																		dps = RhadaTip_computeAllOptimalGear(nthHead, nthNeck, nthShoulder, nthBack, nthChest, nthWrist, nthHands, nthWaist, nthLegs, nthFeet, 
																			nthMainHand, nthSecondaryHand, nthRanged, nthFinger0, nthFinger1, nthTrinket0, nthTrinket1)
																		if tonumber(dps) >= tonumber(topDps) then
																			topDps=dps
																			-- set topDpsSetup to the current config
																			topDpsSetup = (nthHead .. "+" .. nthNeck .. "+" .. nthShoulder .. "+" .. nthBack .. "+" .. nthChest .. "+" .. nthWrist .. "+" .. nthHands .. "+" .. nthWaist .. "+" .. nthLegs .. "+" .. nthFeet .. "+" .. nthMainHand .. "+" .. nthSecondaryHand .. "+" .. nthRanged .. "+" .. nthFinger0 .. "+" .. nthFinger1 .. "+" .. nthTrinket0 .. "+" .. nthTrinket1)
																		end
																		loopNesimo = loopNesimo + 1;
																		-- at this point another full combination of items has been evaluated, before evaluating the next one check if the method should yield. if k is a multiple of slideValue*4, yeld
																		if loopNesimo % (slideValue*4) == 0 then
																			coroutine.yield()
																		end
																	end
																end
															end
														end
													end
												end
											end
										end
									end
								end
							end
						end
					end
				end
			end
		end
	end
	out("The best gear combination is: " .. topDpsSetup)
	-- split topDpsSetup and equip any item that is an actual item link
	local HeadSlot, NeckSlot, ShoulderSlot, BackSlot, ChestSlot, WristSlot, HandsSlot, WaistSlot, LegsSlot, FeetSlot, 
	      MainHandSlot, SecondaryHandSlot, RangedSlot, Finger0Slot, Finger1Slot, Trinket0Slot, Trinket1Slot = strsplit("+", topDpsSetup, 17)
	if strfind(HeadSlot, "Hitem") ~= nil then
		EquipItemByName(HeadSlot, 1)
	end
	if strfind(NeckSlot, "Hitem") ~= nil then
		EquipItemByName(NeckSlot, 2)
	end
	if strfind(ShoulderSlot, "Hitem") ~= nil then
		EquipItemByName(ShoulderSlot, 3)
	end
	if strfind(BackSlot, "Hitem") ~= nil then
		EquipItemByName(BackSlot, 15)
	end
	if strfind(ChestSlot, "Hitem") ~= nil then
		EquipItemByName(ChestSlot, 5)
	end
	if strfind(WristSlot, "Hitem") ~= nil then
		EquipItemByName(WristSlot, 9)
	end
	if strfind(HandsSlot, "Hitem") ~= nil then
		EquipItemByName(HandsSlot, 10)
	end
	if strfind(WaistSlot, "Hitem") ~= nil then
		EquipItemByName(WaistSlot, 6)
	end
	if strfind(LegsSlot, "Hitem") ~= nil then
		EquipItemByName(LegsSlot, 7)
	end
	if strfind(FeetSlot, "Hitem") ~= nil then
		EquipItemByName(FeetSlot, 8)
	end
	if strfind(Finger0Slot, "Hitem") ~= nil then
		EquipItemByName(Finger0Slot, 11)
	end
	if strfind(Finger1Slot, "Hitem") ~= nil then
		EquipItemByName(Finger1Slot, 12)
	end
	if strfind(Trinket0Slot, "Hitem") ~= nil then
		EquipItemByName(Trinket0Slot, 13)
	end
	if strfind(Trinket1Slot, "Hitem") ~= nil then
		EquipItemByName(Trinket1Slot, 14)
	end
	local ItemIdMh;
	if strfind(MainHandSlot, "Hitem") ~= nil then
		ItemIdMh = tonumber(select(3, strfind(MainHandSlot, "Hitem:(%d+)")))
		EquipItemByName(MainHandSlot, 16)
	end
	if strfind(SecondaryHandSlot, "Hitem") ~= nil then
		local ItemIdOh = tonumber(select(3, strfind(SecondaryHandSlot, "Hitem:(%d+)")))
		if ItemIdMh ~= nil and tonumber(ItemIdMh) == tonumber(ItemIdOh) then
			-- search in which bag slot there is still another item like the one already placed into the main hand slot	
			local found = 0
			for bag = 4,0,-1 do
				for slot = GetContainerNumSlots(bag),1,-1 do
					local item = GetContainerItemLink(bag,slot)
					local currentItemID
					if item then
						currentItemID = tonumber(select(3, strfind(item, "Hitem:(%d+)")))
						if currentItemID == ItemIdOh then
							PickupContainerItem(bag,slot) 
							EquipCursorItem(GetInventorySlotInfo("SecondaryHandSlot"));
							--PickupInventoryItem(17)
						end
					end
				end
			end
		else
			-- normal case
			EquipItemByName(SecondaryHandSlot, 17)
		end
	end
	if strfind(RangedSlot, "Hitem") ~= nil then
		EquipItemByName(RangedSlot, 18)
	end
	-- get the end time
	local endTime = time()
	-- and the total time of execution	
	local totalExecutionTime = (endTime)-(RhadaTip_OptimizeGearstartTime)
	Rhada_config_panel_gear_opt_ComputationTime:SetText("RhadaTip has computed and equipped your best gear combination in " .. totalExecutionTime .. " seconds.")
	-- at this point, everything is really over, we can clear everything left behind (putting them to nil should force the garbage collection)
	RhadaTip_ClearGearLists()
	SHOW_OFFSPEC_VALUES = saved_offspec_setting
end


-- this function takes a table of items, calls BuildBonusSet to retrieve the list of eventual tier sets that match the passed item list
-- and returns true or false depending on whether this setCount list is or isnt empty
local function RhadaTip_isItemTier(gearTable)
	for k,v in pairs(gearTable) do
		if rhada_get_tier_number(v) ~= nil then
			return true
		end
	end
	return false
end

-- this function takes an item slot as its input, and computes the highest self dps element for that slot. if this element does not have hit, it puts it into the corresponding list 
-- for that item slot, otherwise it puts in that same list all items (equipped and in the bags) for that slot
-- tier items are dealt like this: if there is at least 1 item among all for that slot that is a tier piece, add them all, otherwise add just the max dps one
local function RhadaTip_GetHeadLinks(slotToCompute)
	-- current dps and max dps values
	local dps, topDps = 0, 0
	local hit = 0
	-- expertise_cap 3.0.0
	local expertise = 0
	local spirit = 0
	local isTierPiece = false
	local highestHeadLink = ""
	-- commented in 3.0.0
	--local bestLinkHasHit = "NO"
	-- expertise_cap 3.0.0
	--local bestLinkHasExpertise = "NO"
	-- for each item in the bags
	for bag = 0,4 do
    		for slot = 1,GetContainerNumSlots(bag) do
      			local item = GetContainerItemLink(bag,slot)
			if item then
	      		local _, _, _, itemLevel, _, _, _, _, itemEquipLoc = GetItemInfo(item)
				if tonumber(itemLevel) >= maxItemLevel then
					-- if the itemEquipLocation of the n-th item in the bags is of interest ...
					if (itemEquipLoc == slotToCompute or 
						(slotToCompute == "INVTYPE_ROBE" and itemEquipLoc == "INVTYPE_CHEST") or 
						(slotToCompute == "INVTYPE_WEAPONMAINHAND" and itemEquipLoc == "INVTYPE_2HWEAPON")  or 
						(slotToCompute == "INVTYPE_WEAPONOFFHAND" and itemEquipLoc == "INVTYPE_HOLDABLE") or 
						(slotToCompute == "INVTYPE_RANGEDRIGHT" and (itemEquipLoc == "INVTYPE_RANGED" or 
						itemEquipLoc == "INVTYPE_THROWN"))) then
						-- set a scanning tooltip up
						RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
						RhadaTipScanningTooltip:SetHyperlink(item)
						-- and get its dps and hit value
						-- expertise_cap 3.0.0
						dps, hit, expertise, spirit = getItemSlotValue()
						-- the next funcion is only called if isTierPiece = "NO" to check if the current item is a tier piece. once it is yes, it means that at least 1 item is a tier piece, so dont update it anymore, it is enough that 1 is to have to add them all
						if not isTierPiece then
							local tempList = {item}
							isTierPiece = RhadaTip_isItemTier(tempList)
						end
						if dps ~= nil then
							if tonumber(dps) > tonumber(topDps) then
								-- update the higher dps
								topDps=dps
								-- and point to the new item
								highestHeadLink = item
							end
						end
						RhadaTipScanningTooltip:Hide()
					end
				end
			end
    		end
  	end
	local currentSlot = ""
	local tableSlot
	-- depending on the slot of interest, set which itemList to use
	if slotToCompute == "INVTYPE_HEAD" then
		tableSlot=RhadaTip_BestHeadLink
		currentSlot = "HeadSlot"
	elseif slotToCompute == "INVTYPE_NECK" then
		tableSlot=RhadaTip_BestNeckLink
		currentSlot = "NeckSlot"
	elseif slotToCompute == "INVTYPE_SHOULDER" then
		tableSlot=RhadaTip_BestShoulderLink
		currentSlot = "ShoulderSlot"
	elseif slotToCompute == "INVTYPE_CLOAK" then
		tableSlot=RhadaTip_BestBackLink
		currentSlot = "BackSlot"
	elseif (slotToCompute == "INVTYPE_ROBE" or slotToCompute == "INVTYPE_CHEST") then
		tableSlot=RhadaTip_BestChestLink
		currentSlot = "ChestSlot"
	elseif slotToCompute == "INVTYPE_WRIST" then
		tableSlot=RhadaTip_BestWristLink
		currentSlot = "WristSlot"
	elseif slotToCompute == "INVTYPE_HAND" then
		tableSlot=RhadaTip_BestHandsLink
		currentSlot = "HandsSlot"
	elseif slotToCompute == "INVTYPE_WAIST" then
		tableSlot=RhadaTip_BestWaistLink
		currentSlot = "WaistSlot"
	elseif slotToCompute == "INVTYPE_LEGS" then
		tableSlot=RhadaTip_BestLegsLink
		currentSlot = "LegsSlot"
	elseif slotToCompute == "INVTYPE_FEET" then
		tableSlot=RhadaTip_BestFeetLink
		currentSlot = "FeetSlot"
	elseif slotToCompute == "INVTYPE_FINGER" then
		tableSlot=RhadaTip_BestFinger0Link
	elseif slotToCompute == "INVTYPE_TRINKET" then
		tableSlot=RhadaTip_BestTrinket0Link
	elseif (slotToCompute == "INVTYPE_WEAPONMAINHAND" or slotToCompute == "INVTYPE_2HWEAPON")then
		tableSlot=RhadaTip_BestMainHandLink
		currentSlot = "MainHandSlot"
	elseif (slotToCompute == "INVTYPE_WEAPONOFFHAND" or slotToCompute == "INVTYPE_HOLDABLE") then
		tableSlot=RhadaTip_BestSecondaryHandLink
		currentSlot = "SecondaryHandSlot"
	elseif (slotToCompute == "INVTYPE_RANGEDRIGHT" or slotToCompute == "INVTYPE_RANGED" or slotToCompute == "INVTYPE_THROWN") then
		tableSlot=RhadaTip_BestRangedLink
		currentSlot = "RangedSlot"
	end
	-- lastly check what the player is using for the slot of interest
	local headLink = GetInventoryItemLink("player",GetInventorySlotInfo(currentSlot))
	-- expertise_cap 3.0.0
	local equippedDps, equippedHit, equippedExpertise = 0, 0, 0
	local equippedIsTierPiece 
	if headLink ~= nil then
		local itemLevel = select(4, GetItemInfo(headLink))
		if tonumber(itemLevel) >= maxItemLevel then
			-- set it to the scanning tooltip
			RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
			RhadaTipScanningTooltip:SetHyperlink(headLink)
			-- get how much dps this item is worth
			-- expertise_cap 3.0.0
			equippedDps, equippedHit, equippedExpertise = getItemSlotValue()
			-- get if it is a tier item (no if needed here since it's one item only)
			--equippedIsTierPiece = RhadaTip_isItemTier()
			local tempList = {headLink}
			equippedIsTierPiece = RhadaTip_isItemTier(tempList)
			--RhadaTip:debug_out(headLink .. " equippedIsTierPiece=" .. equippedIsTierPiece)
			--local tabtemp = {headLink}
			--RhadaTip_cacca(tabtemp)
			RhadaTipScanningTooltip:Hide()
		end
	end
	-- if the highest dps item in the bag is better than the equipped item
	if tonumber(topDps) > tonumber(equippedDps) then
		-- if it doesnt have hit nor expertise nor is a tier item...
		-- if it doesnt have hit nor expertise nor there is a tier item for that slot...
		-- expertise_cap 3.0.0
		--if ( tonumber(expertise) == 0 and tonumber(hit) == 0 and isTierPiece == "NO") then
		if ( tonumber(expertise) == 0 and tonumber(hit) == 0 and isTierPiece == false and equippedIsTierPiece == false) then
			-- insert it into the list
			local itemLevel = select(4, GetItemInfo(highestHeadLink))
			if tonumber(itemLevel) >= maxItemLevel then
				tinsert(tableSlot, highestHeadLink)
			end
		else
			-- otherwise call the RhadaTip_FillSlotList on this item slot to fill the relative list with all items that suit this slot
			RhadaTip_FillSlotList(slotToCompute)
		end
	else
		-- otherwise do the same for the currently equipped item
		-- expertise_cap 3.0.0
		--if ( tonumber(equippedExpertise) == 0 and tonumber(equippedHit) == 0 and not equippedIsTierPiece) then
		if ( tonumber(equippedExpertise) == 0 and tonumber(equippedHit) == 0 and isTierPiece == false and equippedIsTierPiece == false) then
			local itemLevel = select(4, GetItemInfo(headLink))
			if tonumber(itemLevel) >= maxItemLevel then
				tinsert(tableSlot, headLink)
			end
		else
			RhadaTip_FillSlotList(slotToCompute)
		end
	end
end


-- this is used to get the gearDps value obtainable with the combination of items passed as the input
function RhadaTip_computeAllOptimalGear(HeadSlot, NeckSlot, ShoulderSlot, BackSlot, ChestSlot, WristSlot, HandsSlot, WaistSlot, LegsSlot, FeetSlot, MainHandSlot, SecondaryHandSlot, RangedSlot, Finger0Slot, Finger1Slot, Trinket0Slot, Trinket1Slot)
	local spec = GetActiveTalentGroup(false, false)
	RhadaTip.totalGearDps_calc_error = false
	-- the current total gear dps
	local OptimalTotalGearDps = 0
	-- the excessHit rating over the currently selected hit cap, and the current hit rating amount
	local excessHit = 0
	local CaOG_hit_rating = 0
	local CaOG_spirit_rating = 0
	-- expertise_cap 3.0.0
	-- the excessExpertise rating over the currently selected expertise cap, and the current expertise rating amount
	local excessExpertise = 0
	local CaOG_expertiseRating = 0
	-- nth item dps and hit value
	-- expertise_cap 3.0.0
	local temp, tempHit, tempExpertise
	local itemName, ItemLink
	local headLink, neckLink, shoulderLink, backLink, chestLink, wristLink, handsLink, waistLink, legsLink, feetLink, finger0Link, finger1Link, trinket0Link, trinket1Link, mainHandLink, secondaryHandLink, rangedLink
	-- if the passed HeadSlot is not nil and it is an actual itemLink, set it, otherwise leave it at nil
	if HeadSlot ~= nil and strfind(HeadSlot, "Hitem") ~= nil then
		headLink = HeadSlot
	end
	if NeckSlot ~= nil and strfind(NeckSlot, "Hitem") ~= nil then
		neckLink = NeckSlot
	end
	if ShoulderSlot ~= nil and strfind(ShoulderSlot, "Hitem") ~= nil then
		shoulderLink = ShoulderSlot
	end
	if BackSlot ~= nil and strfind(BackSlot, "Hitem") ~= nil then
		backLink = BackSlot
	end
	if ChestSlot ~= nil and strfind(ChestSlot, "Hitem") ~= nil then
		chestLink = ChestSlot
	end
	if WristSlot ~= nil and strfind(WristSlot, "Hitem") ~= nil then
		wristLink = WristSlot
	end
	if HandsSlot ~= nil and strfind(HandsSlot, "Hitem") ~= nil then
		handsLink = HandsSlot
	end
	if WaistSlot ~= nil and strfind(WaistSlot, "Hitem") ~= nil then
		waistLink = WaistSlot
	end
	if LegsSlot ~= nil and strfind(LegsSlot, "Hitem") ~= nil then
		legsLink = LegsSlot
	end
	if FeetSlot ~= nil and strfind(FeetSlot, "Hitem") ~= nil then
		feetLink = FeetSlot
	end
	if Finger0Slot ~= nil and strfind(Finger0Slot, "Hitem") ~= nil then
		finger0Link = Finger0Slot
	end
	if Finger1Slot ~= nil and strfind(Finger1Slot, "Hitem") ~= nil then
		finger1Link = Finger1Slot
	end
	if Trinket0Slot ~= nil and strfind(Trinket0Slot, "Hitem") ~= nil then
		trinket0Link = Trinket0Slot
	end
	if Trinket1Slot ~= nil and strfind(Trinket1Slot, "Hitem") ~= nil then
		trinket1Link = Trinket1Slot
	end
	if MainHandSlot ~= nil and strfind(MainHandSlot, "Hitem") ~= nil then
		mainHandLink = MainHandSlot
	end
	if SecondaryHandSlot ~= nil and strfind(SecondaryHandSlot, "Hitem") ~= nil then
		secondaryHandLink = SecondaryHandSlot
	end
	if RangedSlot ~= nil and strfind(RangedSlot, "Hitem") ~= nil then
		rangedLink = RangedSlot
	end
	-- the following table will be filled with tier items. All items of the 5 possible tier slots that return YES to RhadaTip_isItemTier are inserted into this table
	-- removed the RhadaTip_isItemTier, currently equipped tier items do not return YES as they lose the wording (2) or (4), now that
	--this function isnt used for the optimizer anymore, this would break it
	local gearTable = {}
	-- create a scanning tooltip
	--RhadaTipGemTooltip:SetOwner(owner, "ANCHOR_CURSOR")
	RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	-- if an actual headpiece was passed to the function
	local CaOG_ench_hit = 0
 	if headLink ~= nil then
		-- set it to the scanning tooltip
		RhadaTipScanningTooltip:SetHyperlink(headLink)
		-- the headLink isnt nil, insert it into the gearTable	
		tinsert(gearTable, headLink)
		-- get the dps and hit of the item
		-- expertise_cap 3.0.0 (tempExpertise)
		
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(headLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
 	end
	if neckLink ~= nil then
		RhadaTipScanningTooltip:SetHyperlink(neckLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(neckLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if shoulderLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(shoulderLink)
		tinsert(gearTable, shoulderLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(shoulderLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if backLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(backLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(backLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if chestLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(chestLink)
		tinsert(gearTable, chestLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(chestLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if wristLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(wristLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(wristLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if handsLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(handsLink)
		tinsert(gearTable, handsLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(handsLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if waistLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(waistLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(waistLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if legsLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(legsLink)
		tinsert(gearTable, legsLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(legsLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if feetLink ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(feetLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(feetLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end

 	end
 	if finger0Link ~= nil then
		RhadaTipScanningTooltip:SetHyperlink(finger0Link)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(finger0Link)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if finger1Link ~= nil then
		-- if the finger equipped in the finger1slot isnt the same as the one equipped in the finger0slot, just set the hyperlink
		local finger1Id = tonumber(select(3, strfind(finger1Link, "Hitem:(%d+)")))
		local finger0Id
		if finger0Link ~= nil then
			finger0Id = tonumber(select(3, strfind(finger0Link, "Hitem:(%d+)")))
		end
		if finger1Id ~= finger0Id then
			RhadaTipScanningTooltip:SetHyperlink(finger1Link)
		else
			-- if the finger1 is the same one also in the finger0slot, close and reopen the tooltip
			RhadaTipScanningTooltip:Hide()
			RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
			RhadaTipScanningTooltip:SetHyperlink(finger1Link)
		end
	 	temp, tempHit, tempExpertise = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(finger1Link)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if trinket0Link ~= nil then
	 	RhadaTipScanningTooltip:SetHyperlink(trinket0Link)		
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(trinket0Link)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if trinket1Link ~= nil then
		-- if the trinket equipped in the trinket1slot isnt the same as the one equipped in the trinket0slot, just set the hyperlink
		local trinket1Id = tonumber(select(3, strfind(trinket1Link, "Hitem:(%d+)")))
		local trinket0Id
		if trinket0Link ~= nil then
			  trinket0Id = tonumber(select(3, strfind(trinket0Link, "Hitem:(%d+)")))
		end
		if trinket1Id ~= trinket0Id then
			RhadaTipScanningTooltip:SetHyperlink(trinket1Link)
		else
			-- if the trinket1 is the same one also in the trinket0slot, close and reopen the tooltip
			RhadaTipScanningTooltip:Hide()
			RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
			RhadaTipScanningTooltip:SetHyperlink(trinket1Link)
		end
	 	temp, tempHit, tempExpertise = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(trinket1Link)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if mainHandLink ~= nil then
		RhadaTipScanningTooltip:SetHyperlink(mainHandLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(mainHandLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end
 	if secondaryHandLink ~= nil then
		-- if the weapon equipped in the offhand slot isnt the same as the one equipped in the mainhand slot, just set the hyperlink
		--local itemIdOH = tonumber(secondaryHandLink:match("Hitem:(%d+)"))
		local itemIdOH = tonumber(select(3, strfind(secondaryHandLink, "Hitem:(%d+)")))
		local itemIdMH
		if mainHandLink ~= nil then
	        itemIdMH = tonumber(select(3, strfind(mainHandLink, "Hitem:(%d+)")))
			--itemIdMH = tonumber(mainHandLink:match("Hitem:(%d+)"))
		end
		if itemIdOH ~= itemIdMH then
			RhadaTipScanningTooltip:SetHyperlink(secondaryHandLink)
		else
			-- if the offhand weapon is the same one also in the mainhand slot, close and reopen the tooltip
			RhadaTipScanningTooltip:Hide()
			RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")
			RhadaTipScanningTooltip:SetHyperlink(secondaryHandLink)
		end
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(secondaryHandLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end

 	end
 	if rangedLink ~= nil then
		RhadaTipScanningTooltip:SetHyperlink(rangedLink)
		temp, tempHit, tempExpertise, tempSpirit, tempError = getItemSlotValue(RhadaTipScanningTooltip)
		local enchant_id, enchant_dps, ench_status, enchanted, ench_hit, ench_spirit, ench_exp = RhadaTip_get_enchant_info(rangedLink)
		if not enchanted then
			enchant_dps = 0
			ench_hit = 0
			ench_exp = 0
			ench_spirit = 0
		end

		-- and if neither is nil add them to OptimalTotalGearDps and rating
		CaOG_ench_hit = CaOG_ench_hit + ench_hit
		if temp ~= nil then
			OptimalTotalGearDps = OptimalTotalGearDps + temp
		end
		if tempHit ~= nil then
			CaOG_hit_rating = CaOG_hit_rating + tempHit + ench_hit
		end
		if tempExpertise ~= nil then
			CaOG_expertiseRating = CaOG_expertiseRating + tempExpertise + ench_exp
		end
		if tempSpirit ~= nil then
			CaOG_spirit_rating = CaOG_spirit_rating + tempSpirit +ench_spirit
		end
		if tempError == "Item not loaded" then
			RhadaTip.totalGearDps_calc_error = true
		end
		
 	end

	local exp_rating, expertisecap, base_exp = RhadaTip_class.get_expcap()
--	local effective_exp = CaOG_expertiseRating+base_exp

	if tonumber(CaOG_expertiseRating) > tonumber(expertisecap) then
		excessExpertise = expertisecap-CaOG_expertiseRating
	else 
		excessExpertise = 0
	end	
	
--	out("OptimalTotalGearDps: " .. OptimalTotalGearDps)

	--get current hit rating and hit cap
	local my_hit_rating, CaOG_hitcap, tal_hit_reduction = RhadaTip_class.get_hitcap()
	local total_hit = CaOG_hit_rating + tal_hit_reduction
	if total_hit > CaOG_hitcap then
		-- if the current hit rating is more than the desired hit cap, compute the excessHit
		excessHit = total_hit-CaOG_hitcap
	else 
		excessHit = 0
	end	
-- call the set bonus function
	local tierBonusesDps = RhadaTip:GetSetBonuses(gearTable)

	local post_hit_cap = (excessHit * (RhadaTip_Scales[spec]["HIT_SCALE_POST_CAP"] or 0))
	local pre_hit_cap = (excessHit * (RhadaTip_Scales[spec]["ITEM_MOD_HIT_RATING_SHORT"] or 0))
--	local ench_hit_dps = CaOG_ench_hit * (RhadaTip_Scales[spec]["ITEM_MOD_HIT_RATING_SHORT"] or 0)
	local temp_exp_calc = (excessExpertise*(RhadaTip_Scales[spec]["ITEM_MOD_EXPERTISE_RATING_SHORT"] or 0))
	OptimalTotalGearDps = OptimalTotalGearDps + tierBonusesDps + post_hit_cap - pre_hit_cap - temp_exp_calc
	OptimalTotalGearDps = format("%.2f", OptimalTotalGearDps)
	-- close the scanning tooltip
	RhadaTipScanningTooltip:Hide()
	-- and finally return this gear combination dps value

--	out("my_hit_rating: " .. my_hit_rating)
--	out("CaOG_hit_rating: " .. CaOG_hit_rating)
--	out("CaOG_hitcap: " .. CaOG_hitcap)
--	out("CaOG_ench_hit: " .. CaOG_ench_hit)
--	out("tal_hit_reduction: " .. tal_hit_reduction)
--	out("total_hit: " .. total_hit)
--	out("pre_hit_calc: " .. pre_hit_cap)
--	out("post_hit_calc: " .. post_hit_cap)
--	out("excessHit: " .. excessHit)
--	out("OptimalTotalGearDps: " .. OptimalTotalGearDps)

	return OptimalTotalGearDps
end

-- the coroutine will use this + a local endTime to compute how long it took to finish. making this local will break it
RhadaTip_OptimizeGearstartTime = 0
local saved_offspec_setting = false
-- this function is used to launch the gear optimization. it will launch it either in foreground, or in background, with the appropriate coroutine management. As a consequence of that, all functions in this
-- file and global variables should become local
function RhadaTip_FireGearOptimization(id)
	saved_offspec_setting = SHOW_OFFSPEC_VALUES
	SHOW_OFFSPEC_VALUES = false
	-- get the min itemlevel typed by the user, if any
	if (Rhada_config_panel_gear_opt_InputBox1:GetText() ~= "") then
		-- the user input some value
		maxItemLevel = tonumber(Rhada_config_panel_gear_opt_InputBox1:GetText())
	else
		-- the user didnt input any value, set it high enough to not discard any lvl 80 items
		maxItemLevel = 0
	end
	-- which one of the 2 buttons was pressed?
	--local id = this:GetID()
	local slideValue
	if tonumber(id) == 1601 then
		-- the user pressed foreground, set sliderValue to 20
		slideValue = 20
	else 
		if tonumber(id) == 1602 then
			-- the user pressed background, set sliderValue to the actual Slider1 value
			slideValue = Rhada_config_panel_gear_opt_Slider1:GetValue() 
		end
	end
	-- get the starting time
	RhadaTip_OptimizeGearstartTime = time()
	-- first of all clear all the item_slot lists, so that calling this more than once produces the correct results
	RhadaTip_ClearGearLists()
	-- then initialize a list for each item slot. in this list there will be either 1 element,
	-- if the highest self dps element for that slot doesnt have hit nor is a tier item, or all items for that slot
	RhadaTip_GetHeadLinks("INVTYPE_HEAD");
	RhadaTip_GetHeadLinks("INVTYPE_NECK");
	RhadaTip_GetHeadLinks("INVTYPE_SHOULDER");
	RhadaTip_GetHeadLinks("INVTYPE_CLOAK");
	RhadaTip_GetHeadLinks("INVTYPE_ROBE");
	RhadaTip_GetHeadLinks("INVTYPE_WRIST");
	RhadaTip_GetHeadLinks("INVTYPE_HAND");
	RhadaTip_GetHeadLinks("INVTYPE_WAIST");
	RhadaTip_GetHeadLinks("INVTYPE_LEGS");
	RhadaTip_GetHeadLinks("INVTYPE_FEET");
	-- weapon lists are dealt separately
	RhadaTip_FillWeaponSlotList()
	RhadaTip_GetHeadLinks("INVTYPE_RANGEDRIGHT");
	-- the trinket and ring lists will contain all trinkets and fingers no matter which is the highest dps, if it has hit or not and so on
	RhadaTip_FillSlotList("INVTYPE_FINGER")
	RhadaTip_FillSlotList("INVTYPE_TRINKET")
	-- start the coroutine. assign the function that includes the coroutine to a local coroutine variable, like local co = coroutine.create(foo)	
	local co = coroutine.create(RhadaTip_PrintAllGearSetsLists)
	-- create a temporary frame to time the coroutine execution
	local frame = CreateFrame("Frame",nil,UIParent)
	local counter = 0
	local throttle = 0.2
	frame:SetScript("OnUpdate" , function(self, elapsed)
		counter = counter + elapsed
   			if counter >= throttle then
     				counter = counter - throttle
     				if coroutine.status(co) ~= "dead" then
       					coroutine.resume(co)
   				end
 			end
		end
	)
	-- Actually start the coroutine
 	coroutine.resume(co, slideValue)	
end

-- this function inits the slider (used for the background computation)
function RhadaTip_InitOGEARSlider(slider)
	local sl = getglobal(slider)
	sl.tooltipText = RHADATIP_XML_RT16_SLIDERTOOLTIP; 
	getglobal(slider .. "Low"):SetText("1");
	getglobal(slider .. "High"):SetText("10"); 
	getglobal(slider .. "Text"):SetText("2");
end

-- this function returns a table of the sockets and the quantity of each on the passed item
function RhadaTip_getSocketInfo(tooltip)
	local socketList = {
		Red=0,
		Yellow=0,
		Blue=0,
		Meta=0
	}
	for i=1,tooltip:NumLines() do
		local mytext = getglobal(tooltip:GetName() .. "TextLeft" .. i)
		local text = mytext:GetText()
		if strfind(text, RhadaTip_Localized["Red Socket"]) ~= nil then
			socketList.Red = socketList.Red + 1
		end
		if strfind(text, RhadaTip_Localized["Yellow Socket"]) ~= nil then
			socketList.Yellow = socketList.Yellow + 1
		end
		if strfind(text, RhadaTip_Localized["Blue Socket"]) ~= nil then
			socketList.Blue = socketList.Blue + 1
		end
		if strfind(text, RhadaTip_Localized["Meta Socket"]) ~= nil then
			socketList.Meta = socketList.Meta + 1
		end
	end
	return socketList
end

-- this function computes the item xPS value and returns it together with the hit rating on the item
function getItemSlotValue(passedTooltip, offspec)
	local item_warning = ""
	local tooltip
	local item_is_heirloom = false
	-- this function can be called to compute the dps of items that are either set on RhadaTipScanningTooltip or RhadaTipScanningTooltip2, so it is importat to figure out which one it is
	if RhadaTipScanningTooltip:GetItem() ~= nil then
		tooltip = RhadaTipScanningTooltip
	else 
		tooltip = RhadaTipScanningTooltip2
	end
	-- if this method is called passing a tooltip to it, use that tooltip instead
	if passedTooltip ~= nil then
		tooltip =  passedTooltip
	end
	
	local activeTalentGroup = GetActiveTalentGroup(false, false)
	local spec = activeTalentGroup
	if offspec then
		if activeTalentGroup == 1 then
			spec = 2
		else
			spec = 1
		end
	end
	
	-- get dynamic ItemLink
	local _, ItemLink = tooltip:GetItem();
	if ItemLink == nil then return 0,0,0,0,"","No Item Found on Tip" end
	--check the itemlink to make sure our local item cache is ready
	local p1, p2, p3, p4, p5, p6, p7, p8, p9 = GetItemInfo(ItemLink)
	if p1==nil or p2==nil or p3==nil or p4==nil or p5==nil or p6==nil or p7==nil or p8==nil or p9==nil then return 0,0,0,0,"No Item Found on Tip" end
	--get the item id for reading the cache
	local itemID = rhada_get_item_id(ItemLink)
	
	-- see if we have this item cached
	if RhadaTip.cachedItems[spec] ~= nil then
		if RhadaTip.cachedItems[spec][itemID] ~= nil then
			if RhadaTip.cachedItems[spec][itemID]["dps"] ~= nil then
				return RhadaTip.cachedItems[spec][itemID]["dps"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"], RhadaTip.cachedItems[spec][itemID]["penalty"], RhadaTip.cachedItems[spec][itemID]["warning"]
			end
		end 
	end

	--check to see if this item is equipable. If not, return 0 values.
	local gear_spec, penalty = RhadaTip_class.AP_get_penalty(ItemLink)
	if not gear_spec then 
		RhadaTip.cachedItems[spec] = {[itemID] = {}}
		RhadaTip.cachedItems[spec][itemID]["dps"] = 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"] = 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"] = 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"] = 0
		RhadaTip.cachedItems[spec][itemID]["penalty"] = penalty
		RhadaTip.cachedItems[spec][itemID]["warning"] = "Wrong Proficiency"

		return RhadaTip.cachedItems[spec][itemID]["dps"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"], RhadaTip.cachedItems[spec][itemID]["penalty"], RhadaTip.cachedItems[spec][itemID]["warning"]
	end

	-- itemID is used for RhadaTip_Trinkets(tooltip, itemID)
	local itemID = tonumber(select(3, strfind(ItemLink, "Hitem:(%d+)")))
	-- get static properties
	local _, _, _, _, _, itemType, _, _, itemEquipLoc, _ = GetItemInfo(ItemLink)
	-- init the stat values
	local t_gem_stats = {}

	-- this is needed to activate the socketbonus, based on what the default fill gems are
	local isSocketBonusActive = false
	if (itemEquipLoc == "INVTYPE_TRINKET") then
		-- expertise_cap 3.0.0 now caches extraExpertise too
		local trinketDps, trinketHit, trinketExpertise, trinketSpirit , item_warning = RhadaTip_Trinkets(tooltip, itemID)
--		RhadaTip.cachedItems[spec][itemID] = trinketDps.."|"..trinketHit.."|"..trinketExpertise.."|"..trinketSpirit
		RhadaTip.cachedItems[spec] = {[itemID] = {}}
		RhadaTip.cachedItems[spec][itemID]["dps"] = trinketDps
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"] = trinketHit
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"] = trinketExpertise
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"] = trinketSpirit
		RhadaTip.cachedItems[spec][itemID]["penalty"] = 0
		RhadaTip.cachedItems[spec][itemID]["warning"] = item_warning
--		RhadaTip:debug_out(ItemLink .. " - " .. trinketHit)

		return RhadaTip.cachedItems[spec][itemID]["dps"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"], RhadaTip.cachedItems[spec][itemID]["penalty"], RhadaTip.cachedItems[spec][itemID]["warning"]
	end
	-- do the following only if the item is either a piece of armor or a weapon, excluding trinkets and relics
	if (itemType == RhadaTip_Localized["Armor"] and itemEquipLoc ~= "INVTYPE_TRINKET" and itemEquipLoc ~= "INVTYPE_RELIC") or itemType == RhadaTip_Localized["Weapon"] then
		local stats = {}
		wipe(stats)
		GetItemStats(ItemLink, stats)
		stats = RhadaTip_ReforgeStats(ItemLink, stats)
		
		-- since weapon slots are critial to calculations, we have to determine where this item will be used, then translate
		-- the item raw stats into something our functions can understand. without this context the numbers
		-- cannot be calculated. We still leave the values where they are in the table. We just create new ones with our
		-- expected index values.
		if (itemType == RhadaTip_Localized["Weapon"] and (itemEquipLoc == "INVTYPE_2HWEAPON" or itemEquipLoc == "INVTYPE_WEAPONMAINHAND" or itemEquipLoc == "INVTYPE_WEAPON")) then
			stats["mainhand"] = format("%.1f", stats["ITEM_MOD_DAMAGE_PER_SECOND_SHORT"])
		elseif (itemType == RhadaTip_Localized["Weapon"] and (itemEquipLoc == "INVTYPE_WEAPONOFFHAND" or itemEquipLoc == "INVTYPE_RANGED")) then 
			stats["offhand"] = format("%.1f", stats["ITEM_MOD_DAMAGE_PER_SECOND_SHORT"])
		end

		-- TipTac work around for broken/halfloaded tooltips. Since the real ones get funky, we'll use our own.
		if (IsAddOnLoaded("TipTac")) then
			RhadaTipScanningTooltip2:SetOwner(UIParent, "ANCHOR_NONE")
			RhadaTipScanningTooltip2:SetHyperlink(ItemLink)
			
			tooltip = RhadaTipScanningTooltip2
		end			

--		check to see if our localizations have failed. If so, skip the socket bonus and default gem code.
		-- THIS LINE IS NOT TESTING FOR THE BONUS!!!. It is simply testing to see if the value is loaded in out localizations.
--		if RhadaTip_Localized["Socket Bonus"] ~= nil and not (IsAddOnLoaded("TipTac")) then
		if RhadaTip_Localized["Socket Bonus"] ~= nil then
			-- for each line in the tooltip do the following
			for i=1,tooltip:NumLines() do
				-- read the nth line
				local mytext = getglobal(tooltip:GetName() .. "TextLeft" .. i)
				-- get the colour of the n-th line
				local r,g,b = mytext:GetTextColor()
				-- get the text out of it
				local text = mytext:GetText()
				if text == "Retrieving item information" or text == nil then
					return 0,0,0,0,penalty, "Item not loaded"
				end
				
				-- this section is used to simulate gems in ungemmed items.
				-- this is for the metagem only, so no socketbonus is used
				isSocketBonusActive = false
				if strfind(text, RhadaTip_Localized["Meta Socket"]) ~= nil then
					if RHAD_META_GEM ~= nil then
						-- this means that at least one empty socket has been found
						oneEmptySocket = 0
						local gemId = rhada_get_item_id(RHAD_META_GEM)
						-- call RhadaTip_IsGemMeta to decide if the empty Meta Socket is actually filled with a meta gem
						-- if it is, isSocketBonusActive will remain 0, otherwise it will start to grow
						if rhada_gem_is_meta(tonumber(gemId)) then
							isSocketBonusActive = true
						end

						local gem_stats = Rhada_Gem_Data_Retrieve(tonumber(gemId))
						for k,v in pairs(stats) do
							t_gem_stats[k] = (t_gem_stats[k] or 0) + (gem_stats[k] or 0)
						end
					end
				end
				-- if the line is an empty Red Socket
				if strfind(text, RhadaTip_Localized["Red Socket"]) ~= nil then
					-- check if the user has selected a default red filling gem
					if RHAD_RED_GEM ~= nil then
						--get our red default gem
						local gemId = rhada_get_item_id(RHAD_RED_GEM)
						--check to see if it qualifies as red
						if rhada_gem_is_red(gemId) then
							--if it does not, then flag the bonus as invalid
							isSocketBonusActive = true
						end
						--get the stats for our gem
						local gem_stats = Rhada_Gem_Data_Retrieve(tonumber(gemId))
						
						--log the gem stats for our dps calculations
						for k,v in pairs(stats) do
							t_gem_stats[k] = (t_gem_stats[k] or 0) + (gem_stats[k] or 0)
						end
						item_warning = "Using Default Gems"
					end
				end
				if strfind(text, RhadaTip_Localized["Yellow Socket"]) ~= nil then
					-- check if the user has selected a default yellow filling gem
					if RHAD_YELLOW_GEM ~= nil then
						--get our yellow default gem
						local gemId = rhada_get_item_id(RHAD_YELLOW_GEM)
						--check to see if it qualifies as yellow
						if rhada_gem_is_yellow(gemId) then
							--if it does not, then flag the bonus as invalid
							isSocketBonusActive = true
						end
						--get the stats for our gem
						local gem_stats = Rhada_Gem_Data_Retrieve(tonumber(gemId))
						
						--log the gem stats for our dps calculations
						for k,v in pairs(stats) do
							t_gem_stats[k] = (t_gem_stats[k] or 0) + (gem_stats[k] or 0)
						end
						item_warning = "Using Default Gems"
					end
				end
				if strfind(text, RhadaTip_Localized["Blue Socket"]) ~= nil then
					-- check if the user has selected a default blue filling gem
					if RHAD_BLUE_GEM ~= nil then
						--get our blue default gem
						local gemId = rhada_get_item_id(RHAD_BLUE_GEM)
						--check to see if it qualifies as blue
						if rhada_gem_is_blue(gemId) then
							--if it does not, then flag the bonus as invalid
							isSocketBonusActive = true
						end
						--get the stats for our gem
						local gem_stats = Rhada_Gem_Data_Retrieve(tonumber(gemId))
						
						--log the gem stats for our dps calculations
						for k,v in pairs(stats) do
							t_gem_stats[k] = (t_gem_stats[k] or 0) + (gem_stats[k] or 0)
						end
						item_warning = "Using Default Gems"
					end
				end
				
				
				if strfind(text, RhadaTip_Localized["Socket Bonus"]) ~= nil then
					-- the current line has the Socket Bonus information, first check if the socket bonus is active, to do so check either if the line is written in green, or if isSocketBonusActive is still 0 (as soon as a wrong colour gem fills a unsocketed slot, it starts to grow)
					if r+b == 0 or isSocketBonusActive then
						-- if the socket bonus gives stamina save the value into socketBonusStamina
						if strfind(text, RhadaTip_Localized["Stamina"]) ~= nil then
							stats["ITEM_MOD_STAMINA_SHORT"] = (stats["ITEM_MOD_STAMINA_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["Stamina"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["Intellect"]) ~= nil then
							stats["ITEM_MOD_INTELLECT_SHORT"] = (stats["ITEM_MOD_INTELLECT_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["Intellect"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["Spirit"]) ~= nil then
							stats["ITEM_MOD_SPIRIT_SHORT"] = (stats["ITEM_MOD_SPIRIT_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["Spirit"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["Strength"]) ~= nil then
							stats["ITEM_MOD_STRENGTH_SHORT"] = (stats["ITEM_MOD_STRENGTH_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["Strength"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["Agility"]) ~= nil then
							stats["ITEM_MOD_AGILITY_SHORT"] = (stats["ITEM_MOD_AGILITY_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["Agility"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["MasteryRating"]) ~= nil then
							stats["ITEM_MOD_MASTERY_RATING_SHORT"] = (stats["ITEM_MOD_MASTERY_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["MasteryRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["ExpertiseRating"]) ~= nil then
							stats["ITEM_MOD_EXPERTISE_RATING_SHORT"] = (stats["ITEM_MOD_EXPERTISE_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["ExpertiseRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["SpellPower"]) ~= nil then
							stats["ITEM_MOD_SPELL_POWER_SHORT"] = (stats["ITEM_MOD_SPELL_POWER_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["SpellPower"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["AttackPower"]) ~= nil then
							stats["ITEM_MOD_ATTACK_POWER_SHORT"] = (stats["ITEM_MOD_ATTACK_POWER_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["AttackPower"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["HitRating"]) ~= nil then
							stats["ITEM_MOD_HIT_RATING_SHORT"] = (stats["ITEM_MOD_HIT_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["HitRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["CritRating"]) ~= nil then
							stats["ITEM_MOD_CRIT_RATING_SHORT"] = (stats["ITEM_MOD_CRIT_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["CritRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["HasteRating"]) ~= nil then
							stats["ITEM_MOD_HASTE_RATING_SHORT"] = (stats["ITEM_MOD_HASTE_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["HasteRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["DodgeRating"]) ~= nil then
							stats["ITEM_MOD_DODGE_RATING_SHORT"] = (stats["ITEM_MOD_DODGE_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["DodgeRating"])) or 0)
						end
						if strfind(text, RhadaTip_Localized["ParryRating"]) ~= nil then
							stats["ITEM_MOD_PARRY_RATING_SHORT"] = (stats["ITEM_MOD_PARRY_RATING_SHORT"] or 0) + (select(3, strfind(text, "%+(%d+) ".. RhadaTip_Localized["ParryRating"])) or 0)
						end
					end
				end				
			end
		else
			item_warning = "RT:Socket Bonus/Emtpy Gems not Calculated"
		end
		
		--hide our scanning tip if we used it above for tiptac
		if (IsAddOnLoaded("TipTac")) then
			RhadaTipScanningTooltip2:Hide()
		end
		
		-- this section is used to determine what gems are socketed into the item, and determine dps
		-- contributions.
		local gem_stats = RhadaTip_GetItemGemStats(ItemLink)
		
--		local gem_socket_count = 0
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_BLUE"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_COGWHEEL"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_HYDRAULIC"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_META"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_NO_COLOR"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_PRISMATIC"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_RED"] or 0)
--		gem_socket_count = gem_socket_count + (stats["EMPTY_SOCKET_YELLOW"] or 0)
--		out(ItemLink .. ":" .. gem_socket_count)
		
		if t_gem_stats ~= nil then
			for k,v in pairs(t_gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end

		if gem_stats ~= nil then
			for k,v in pairs(gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end

		local scaled_stats = 0
		for k,v in pairs(stats) do
			scaled_stats = scaled_stats + (v * (RhadaTip_Scales[spec][k] or 0))
		end
		
		--calculate our item procs if any
		local proc_dps = 0
		if RhadaTip_gear_procs[itemID] ~= nil then
			local activeTalentGroup = GetActiveTalentGroup(false, false)
			for k,v in pairs(RhadaTip_gear_procs[itemID]) do
				local t_proc = onProcUptime(v["proc_len"], v["proc_chan"], v["proc_ICD"])
				--take the proc rate and multiply it times our scale value
				proc_dps = proc_dps + t_proc * v["stat_val"] * RhadaTip_Scales[activeTalentGroup][v["proc_stat"]]
			end
		end

		local gain = scaled_stats+proc_dps

		gain = format("%.2f", gain)
		-- expertise_cap 3.0.0 now caches extraExpertise too
--		RhadaTip.cachedItems[spec][itemID] = gain.."|"..extraHit.."|"..extraExpertise.."|"..extraSpirit
		RhadaTip.cachedItems[spec] = {[itemID] = {}}
		RhadaTip.cachedItems[spec][itemID]["dps"] = gain - penalty
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"] = stats["ITEM_MOD_HIT_RATING_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"] = stats["ITEM_MOD_EXPERTISE_RATING_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"] = stats["ITEM_MOD_SPIRIT_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["penalty"] = penalty
		RhadaTip.cachedItems[spec][itemID]["warning"] = item_warning
--		RhadaTip:debug_out(ItemLink .. ":" .. gain.."|"..extraHit.."|"..extraExpertise)
		-- expertise_cap 3.0.0 now returns extraExpertise too
		return RhadaTip.cachedItems[spec][itemID]["dps"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"], RhadaTip.cachedItems[spec][itemID]["penalty"], RhadaTip.cachedItems[spec][itemID]["warning"]
	end
	-- to stop having issues with relics, return 0 dps and 0 hit for them
	if (itemEquipLoc == "INVTYPE_RELIC") then
		local activeTalentGroup = GetActiveTalentGroup(false, false)
		local stats = {}
		wipe(stats)
		GetItemStats(ItemLink, stats)
		stats = RhadaTip_ReforgeStats(ItemLink, stats)
		local stat_dps = 0
		--loop through the stats and multiply them by our scale values.
		for stat, value in pairs(stats) do
			if RhadaTip_Scales[activeTalentGroup][stat] ~= nil then
				stat_dps = stat_dps + RhadaTip_Scales[activeTalentGroup][stat] * value
--				RhadaTip:debug_out(stat .. ":" .. value)
			end		
		end
		-- expertise_cap 3.0.0 now returns extraExpertise too
		RhadaTip.cachedItems[spec] = {[itemID] = {}}
		RhadaTip.cachedItems[spec][itemID]["dps"] = stat_dps
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"] = stats["ITEM_MOD_HIT_RATING_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"] = stats["ITEM_MOD_EXPERTISE_RATING_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"] = stats["ITEM_MOD_SPIRIT_SHORT"] or 0
		RhadaTip.cachedItems[spec][itemID]["penalty"] = 0
		RhadaTip.cachedItems[spec][itemID]["warning"] = ""

		return RhadaTip.cachedItems[spec][itemID]["dps"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_HIT_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_EXPERTISE_RATING_SHORT"], RhadaTip.cachedItems[spec][itemID]["ITEM_MOD_SPIRIT_SHORT"], RhadaTip.cachedItems[spec][itemID]["penalty"], RhadaTip.cachedItems[spec][itemID]["warning"]
	end
end


function RhadaTip_get_equip_stat(itemEquipLoc, stat)
	local ItemLink = GetInventoryItemLink("player",GetInventorySlotInfo(RhadaTip_EquipSlotTranslate[itemEquipLoc]))
	if ItemLink == nil then return 0 end
	local stats = {}
	wipe(stats)
	GetItemStats(ItemLink, stats)
	stats = RhadaTip_ReforgeStats(ItemLink, stats)
	return stats[stat] or 0
end

function rhada_gear_lvl()
	local headLink = GetInventoryItemLink("player",GetInventorySlotInfo("HeadSlot"))
	local neckLink = GetInventoryItemLink("player",GetInventorySlotInfo("NeckSlot"))
	local shoulderLink = GetInventoryItemLink("player",GetInventorySlotInfo("ShoulderSlot"))
	local backLink = GetInventoryItemLink("player",GetInventorySlotInfo("BackSlot"))
	local chestLink = GetInventoryItemLink("player",GetInventorySlotInfo("ChestSlot"))
	local wristLink = GetInventoryItemLink("player",GetInventorySlotInfo("wristSlot"))
	local handsLink = GetInventoryItemLink("player",GetInventorySlotInfo("HandsSlot"))
	local waistLink = GetInventoryItemLink("player",GetInventorySlotInfo("WaistSlot"))
	local legsLink = GetInventoryItemLink("player",GetInventorySlotInfo("LegsSlot"))
	local feetLink = GetInventoryItemLink("player",GetInventorySlotInfo("FeetSlot"))
	local finger0Link = GetInventoryItemLink("player",GetInventorySlotInfo("Finger0Slot"))
	local finger1Link = GetInventoryItemLink("player",GetInventorySlotInfo("Finger1Slot"))
	local trinket0Link = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket0Slot"))
	local trinket1Link = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket1Slot"))
	local mainHandLink = GetInventoryItemLink("player",GetInventorySlotInfo("MainHandSlot"))
	local secondaryHandLink = GetInventoryItemLink("player",GetInventorySlotInfo("SecondaryHandSlot"))
	local rangedLink = GetInventoryItemLink("player",GetInventorySlotInfo("RangedSlot"))
	local cur_item_lvl_total = 0
	local item_count = 0

	if headLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(headLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if neckLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(neckLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if shoulderLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(shoulderLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if backLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(backLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if chestLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(chestLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if wristLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(wristLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if handsLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(handsLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if waistLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(waistLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if legsLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(legsLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if feetLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(feetLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if finger0Link ~= nil then
		local _, _, _, ilvl = GetItemInfo(finger0Link)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if finger1Link ~= nil then
		local _, _, _, ilvl = GetItemInfo(finger1Link)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if trinket0Link ~= nil then
		local _, _, _, ilvl = GetItemInfo(trinket0Link)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if trinket1Link ~= nil then
		local _, _, _, ilvl = GetItemInfo(trinket1Link)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if mainHandLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(mainHandLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if secondaryHandLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(secondaryHandLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end
	if rangedLink ~= nil then
		local _, _, _, ilvl = GetItemInfo(rangedLink)
		if ilvl ~= nil then
			cur_item_lvl_total = cur_item_lvl_total + ilvl
			item_count = item_count + 1
		end
	end	
	cur_item_lvl_total = cur_item_lvl_total/item_count
	return cur_item_lvl_total
end

function rhada_get_item_id(ItemLink)
	if ItemLink == nil then return 0 end
	local itemID = tonumber(select(3, strfind(ItemLink, "Hitem:(%d+)")))
	return itemID
end

function rhada_get_enchant_id(ItemLink)
	if ItemLink == nil then return "" end
	local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId, linkLevel, reforgeId = strsplit(":", ItemLink)
	return enchantId
end

function RhadaTip_GetItemGemStats(ItemLink)
	--get gem stats
	if ItemLink == nil then return end
	local stats = {}
	local gem_stats = {}
	local spec = GetActiveTalentGroup(false, false)
	local _, gemLink = GetItemGem(ItemLink,1)
	local jewelId1 = rhada_get_item_id(gemLink)
	if jewelId1 > 0  then
		gem_stats = Rhada_Gem_Data_Retrieve(jewelId1)
		if gem_stats ~= nil then
			for k,v in pairs(gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end
	end
	local _, gemLink = GetItemGem(ItemLink,2)
	local jewelId2 = rhada_get_item_id(gemLink)
	if jewelId2 > 0  then
		gem_stats = Rhada_Gem_Data_Retrieve(jewelId2)
		if gem_stats ~= nil then
			for k,v in pairs(gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end
	end
	local _, gemLink = GetItemGem(ItemLink,3)
	local jewelId3 = rhada_get_item_id(gemLink)
	if jewelId3 > 0  then
		gem_stats = Rhada_Gem_Data_Retrieve(jewelId3)
		if gem_stats ~= nil then
			for k,v in pairs(gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end
	end
	local _, gemLink = GetItemGem(ItemLink,4)
	local jewelId4 = rhada_get_item_id(gemLink)
	if jewelId4 > 0  then
		gem_stats = Rhada_Gem_Data_Retrieve(jewelId4)
		if gem_stats ~= nil then
			for k,v in pairs(gem_stats) do
				if RhadaTip_Scales[spec][k] then
					stats[k] = (stats[k] or 0) + v
				end
			end
		end
	end
	return stats
end

function RhadaTip_GetInstanceBossList()

	local zone, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
	zone = strlower(gsub(zone, " ", ""))
	if zone == "lostcityofthetol'vir" then
		zone =  "lostcityoftolvir"
	end
	wipe(RhadaTip_InstGearUpgrades)
	RhadaTip_InstGearUpgrades = {}
	if RhadaTip_Inst_List ~= nil and RhadaTip_Inst_List[zone] ~= nil then
--		out("Found Instance")
		for k,v in pairs(RhadaTip_Inst_List[zone]) do
--			gear_list = RhadaTip_FindInstanceUpgrades(RhadaTip_Inst_List[zone], RhadaTip_Inst_type[difficultyIndex])
			RhadaTip_FindInstanceUpgrades(v, RhadaTip_Inst_type[difficultyIndex])
		end
	end
	out("Possible " .. zone .. " Upgrades:")
	if RhadaTip_InstGearUpgrades then
		local item_count = 0
		for k,v in pairs(RhadaTip_InstGearUpgrades) do
			if item_count < 10 then
				DEFAULT_CHAT_FRAME:AddMessage(v.item_dps .."-".. v.item .."(".. v.boss .. "/" .. (v.instance or "") ..")")
				item_count = item_count +1
			end
		end
	end

end

function RhadaTip_FindInstanceUpgrades(zone, Inst_type)
--	out(zone .. ":" .. Inst_type)
	RhadaTipScanningTooltip:SetOwner(UIParent, "ANCHOR_NONE")

	for a,b in pairs(AtlasLoot_Data[zone]) do
		if a == "Normal" or a == "Heroic" then
--			out(a)
			--step through pages
			for c,d in pairs(b) do
				--step through items
				-- now we go look for an upgrade.
				for e,f in pairs(d) do
					if type(f[2]) == "table" then
					--there are items out there that have multiple faction subsets. 
					--atlas loot groups these into a subtable. Eventually we need to parse these as well
					else
						--make sure the pointer is valid
						if f[2] then
							--check to see if we have a valid item id
							if f[2] ~= "0" and f[2] ~= 0 then
								--get the item stats for the item in the database
--								out(c .. ":" .. e .. ":" .. f[2])
								local iname, itemLink, _, _, _, itemType, itemSubType, _, itemEquipLoc, _ = GetItemInfo(f[2])

								if itemLink ~= nil and (itemType == RhadaTip_Localized["Armor"] or itemType == RhadaTip_Localized["Weapon"]) then

									--load the item into our scanning tooltip
									RhadaTipScanningTooltip:SetHyperlink(itemLink)
									--calculate the item's dps
									local item_dps = getItemSlotValue(RhadaTipScanningTooltip, false)
									item_dps = tonumber(item_dps)
--									out(itemLink .. ":" .. item_dps)
									
									local my_item_dps = 0
									local mylink = GetInventoryItemLink("player",GetInventorySlotInfo(RhadaTip_EquipSlotTranslate[itemEquipLoc]))
									--load the item into our scanning tooltip
									if mylink ~= nil then
										RhadaTipScanningTooltip:SetHyperlink(mylink)
										--calculate the item's dps
										my_item_dps = getItemSlotValue(RhadaTipScanningTooltip, false)
										my_item_dps = tonumber(my_item_dps)
--										out(mylink .. ":" .. my_item_dps)
										--check to see if we have a possible upgrade
										if item_dps > my_item_dps then
											tinsert(RhadaTip_InstGearUpgrades, {item = itemLink, item_type = itemType, slot = itemEquipLoc, item_dps = tonumber(item_dps), boss = AtlasLoot_Data[zone].info.name, instance = AtlasLoot_Data[zone].info.instance})
										end
									end
								end
							end
						end
					end
				end
			end
		end
 		
	end
	RhadaTipScanningTooltip:Hide()
end
