local SPI = 1
local DODGE = 2
local PARRY = 3
local HIT = 4
local CRIT = 5
local HASTE = 6
local EXP = 7
local MASTERY = 8
-- "ITEM_MOD_STAMINA_SHORT" 
-- "ITEM_MOD_INTELLECT_SHORT" 
-- "ITEM_MOD_SPIRIT_SHORT" 
-- "ITEM_MOD_STRENGTH_SHORT" 
-- "ITEM_MOD_AGILITY_SHORT" 
-- "ITEM_MOD_HIT_RATING_SHORT" 
-- "ITEM_MOD_HASTE_RATING_SHORT" 
-- "ITEM_MOD_SPELL_POWER_SHORT" 
-- "ITEM_MOD_CRIT_RATING_SHORT" 
-- "ITEM_MOD_ATTACK_POWER_SHORT" 
-- "ITEM_MOD_MASTERY_RATING_SHORT" 
-- "ITEM_MOD_EXPERTISE_RATING_SHORT" 
-- "ITEM_MOD_DODGE_RATING_SHORT" 
-- "ITEM_MOD_PARRY_RATING_SHORT" 

-- Array of long stat names. Index = number returned by GetReforgedStatIDs.
rhada_StatNames = {
	"ITEM_MOD_SPIRIT_SHORT",
	"ITEM_MOD_DODGE_RATING_SHORT",
	"ITEM_MOD_PARRY_RATING_SHORT",
	"ITEM_MOD_HIT_RATING_SHORT",
	"ITEM_MOD_CRIT_RATING_SHORT",
	"ITEM_MOD_HASTE_RATING_SHORT",
	"ITEM_MOD_EXPERTISE_RATING_SHORT",
	"ITEM_MOD_MASTERY_RATING_SHORT"
}

-- Array of short stat names. Index = number returned by GetReforgedStatIDs.
rhada_ShortStatNames = {
	"ITEM_MOD_SPIRIT_SHORT",
	"ITEM_MOD_DODGE_RATING_SHORT",
	"ITEM_MOD_PARRY_RATING_SHORT",
	"ITEM_MOD_HIT_RATING_SHORT",
	"ITEM_MOD_CRIT_RATING_SHORT",
	"ITEM_MOD_HASTE_RATING_SHORT",
	"ITEM_MOD_EXPERTISE_RATING_SHORT",
	"ITEM_MOD_MASTERY_RATING_SHORT"
}

-- List of all possible reforgings.
-- First entry has id <offset>, so the order is important!
-- Each entry has the form {decreased stat, increased stat}
local rhada_reforgeIDs = {
	{SPI, DODGE},
	{SPI, PARRY},
	{SPI, HIT},
	{SPI, CRIT},
	{SPI, HASTE},
	{SPI, EXP},
	{SPI, MASTERY},
	
	{DODGE, SPI},
	{DODGE, PARRY},
	{DODGE, HIT},
	{DODGE, CRIT},
	{DODGE, HASTE},
	{DODGE, EXP},
	{DODGE, MASTERY},
	
	{PARRY, SPI},
	{PARRY, DODGE},
	{PARRY, HIT},
	{PARRY, CRIT},
	{PARRY, HASTE},
	{PARRY, EXP},
	{PARRY, MASTERY},
	
	{HIT, SPI},
	{HIT, DODGE},
	{HIT, PARRY},
	{HIT, CRIT},
	{HIT, HASTE},
	{HIT, EXP},
	{HIT, MASTERY},
	
	{CRIT, SPI},
	{CRIT, DODGE},
	{CRIT, PARRY},
	{CRIT, HIT},
	{CRIT, HASTE},
	{CRIT, EXP},
	{CRIT, MASTERY},
	
	{HASTE, SPI},
	{HASTE, DODGE},
	{HASTE, PARRY},
	{HASTE, HIT},
	{HASTE, CRIT},
	{HASTE, EXP},
	{HASTE, MASTERY},
	
	{EXP, SPI},
	{EXP, DODGE},
	{EXP, PARRY},
	{EXP, HIT},
	{EXP, CRIT},
	{EXP, HASTE},
	{EXP, MASTERY},
	
	{MASTERY, SPI},
	{MASTERY, DODGE},
	{MASTERY, PARRY},
	{MASTERY, HIT},
	{MASTERY, CRIT},
	{MASTERY, HASTE},
	{MASTERY, EXP},
}

function rhada_GetReforgeID(itemLink)
	if itemLink ~= nil then
		local _, _, debug_item = strfind(itemLink, "^|c%x+|H(.+)|h%[.*%]")
		local p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11 = strsplit(":", debug_item)
		return (p11 ~= "0") and p11 or nil
	end
	return "0"
end

function reforge_val(itemLink)
	local reforge_error = false
	local rf_stat = {}

	if rhada_IsItemReforged(itemLink) then
		local rf_ID = rhada_GetReforgeID(itemLink)		
		if (rf_ID > "100") and (rf_ID ~= nil) then
			local minus, plus = rhada_GetReforgedStatIDs(rf_ID)
			if minus and plus then
				rf_stat[rhada_ShortStatNames[minus]] = rhada_ShortStatNames[plus]
			else
				reforge_error = true
			end
		else
			reforge_error = true
		end
	end	
	return reforge_error, rf_stat
end

function rhada_GetReforgedStatIDs(id)
	local rid = id - 113 + 1
	local info = rhada_reforgeIDs[rid]
	if info then
--		RhadaTip:debug_out("rhada_GetReforgedStatIDs " .. info[1] .. ":" .. info[2])
		return info[1], info[2]
	else
		return "", ""
	end
end

function rhada_IsItemReforged(itemLink)
	local id = rhada_GetReforgeID(itemLink)
	return id ~= nil and (id > "100")
end

function RhadaTip_ReforgeStats(ItemLink, stats)
	if rhada_IsItemReforged(ItemLink) then
		--our pair of reforge stats
		local rf_stats = {}
		local rf_error = false
		--get the reforge pair for this item
		rf_error, rf_stats = reforge_val(ItemLink)
		--make sure there is no error retrieving them
		if not rf_error then
			for k,v in pairs(rf_stats) do
				--store the diminished value into our secondary stat slot. We DO NOT ADD in this
				--case because the stats CANNOT exist by definition. Reforge only allows us to
				--move stats to new stats that do not exist.
				stats[v] = math.floor(stats[k] * .4)
				
				--since we have calculated our new stat value, we can reduce or original stat.
				stats[k] = math.ceil(stats[k] * .6)
			end
		end
	end
	return stats
end